#define INCL_DOS
#include <os2.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "..\..\usbcalls.h"

#define BULK_IOCTL 1

#pragma pack(1)

typedef struct _SCSICDB6 { /* CDB6 - 6 byte Command Descriptor Block*/

        UCHAR   Opcode;             /* CDB Operation Code           */
        UCHAR   Lun_MsbLBA;         /* SCSI LUN & 5 MSB bits of LBA */
        UCHAR   MidLBA;             /* SCSI MID byte of LBA         */
        UCHAR   LsbLBA;             /* SCSI LBA byte of LBA         */
        UCHAR   XferLen;            /* SCSI Xfer length, Alloc length */
        UCHAR   Control;            /* Control byte                 */

} SCSICDB6;

typedef struct _SCSICDB10 { /* CDB10 - 10 byte Command Descriptor Block*/

        UCHAR   Opcode;            /* CDB Operation Code           */
        UCHAR   Lun;               /* SCSI                         */
        UCHAR   LBA[4];            /* SCSI LBA MSB->LSB            */
        UCHAR   Res;               /* reserved byte                */
        UCHAR   XferLen[2];        /* SCSI Xfer length (MSB first) */
        UCHAR   Control;           /* Control byte                 */

} SCSICDB10;

typedef struct _SCSICDB12 { /* CDB12 - 12 byte Command Descriptor Block*/


        UCHAR   Opcode;            /* CDB Operation Code           */
        UCHAR   Lun;               /* SCSI                         */
        UCHAR   LBA[4];            /* SCSI LBA MSB->LSB            */
        UCHAR   XferLen[4];        /* SCSI Xfer length (MSB first) */
        UCHAR   Res;               /* reserved byte                */
        UCHAR   Control;           /* Control byte                 */

} SCSICDB12;

#define  CBW_SIGNATURE  0x43425355L // = "USBC", Command Block Wrapper Signature
#define  CBW_TAG        0x89AA2008L
#define  CBW_FLAGS_IN         0x80  // Direction: data in
#define  CBW_LUN              0x00
#define  CB_MAX_LENGTH          16  // Command Block Max Length


typedef  struct _CBWrapper
{                                // byte
   ULONG    dSignature;          // 0  = 0x43425355L = "USBC"
   ULONG    dTag;                // 4  is not interpreted or altered by the MSD and
                                 //    echoed by the MSD in the CSW
   ULONG    dDataTrLength;       // 8  in bytes; if is set to 0, no data will be transferred
   UCHAR    bFlags;              // C  Bit7-Direction: 0 = data out, 1 = data in; Bits 6...0 reserved
   UCHAR    bLUN;                // D  Logical Unit Number
   UCHAR    bCBLength;           // E  valid Length of the Command Block
   union
   {
     UCHAR     cb[CB_MAX_LENGTH];// F  Command Block to be executed by the device
     SCSICDB6  cdb6;
     SCSICDB10 cdb10;
     SCSICDB12 cdb12;
   };

}  CBWrapper;

//       Command Status Wrapper

#define  CSW_SIGNATURE           0x53425355L // = "USBS", Command Status Wrapper Signature

typedef  struct _CSWrapper
{                          // byte
   ULONG    dSignature;    // 0  = 0x53425355L = "USBS"
   ULONG    dTag;          // 4  the MSD shall echo this value as received in the CBW
   ULONG    dDataResidue;  // 8  dDataTrLength - the actual amount transferred to/from the MSD
   UCHAR    bStatus;       // C  indicates the success or failure of the command

}  CSWrapper;



#define SCSI_INQUERY         0x12
#define INQUERY_LEN          0x24

#define SCSI_FORMAT_CAPACITY 0x23
#define FORMAT_CAPACITY_LEN  0x40

#define SCSI_READ_CAPACITY   0x25
#define READ_CAPACITY_LEN    0x08

#define SCSI_READ10          0x28
#define SECTOR_SIZE          0x0200

#define SCSI_START_STOP      0x00


#pragma pack()

void DumpData(UCHAR *ucData, ULONG ulActual)
{
    ULONG i;
    printf("Data : ");
    for( i=0; i<ulActual; i++)
    {
      printf("%02X ",ucData[i]);
      if(((i+1)%16)==0)
        printf("\r\n       ");
    }
    printf("\r\n");
}

int scsi_cmd(int fd, unsigned char cdb[], int clen, scsi_io_mode_t mode,
	     void *data, size_t len, void *extra_data)
{
  
}

APIRET issue_SCSI(USBHANDLE hUSBDev, struct _CBWrapper *CBW, UCHAR *ucData, ULONG *ulActual )
{
  struct _CSWrapper CSW;
  ULONG  ulReadCSW;
  APIRET rc;

  rc = UsbBulkWrite( hUSBDev, 0x01, 0x00, sizeof(struct _CBWrapper), (UCHAR *)CBW, 16);
  if(rc)
  {
    printf("Failed to write CBW rc=%d\n",rc);
    return rc;
  }
  if( *ulActual )
  {
    if( CBW->bFlags & CBW_FLAGS_IN)
    {
      rc = UsbBulkRead(hUSBDev, 0x81, 0x00, ulActual, ucData, 1000);
    } else
    {
      rc = UsbBulkWrite(hUSBDev, 0x01, 0x00, *ulActual, ucData, 1000);
    }
    if(rc & (rc != USB_ERROR_LESSTRANSFERED) )
    {
      printf("Failed to Send/Receive Data rc=%d\n",rc);
      return rc;
    }
    printf("Read/Write %d bytes\n", *ulActual);
  }
  ulReadCSW= sizeof(CSW);
  rc = UsbBulkRead(hUSBDev, 0x81, 0x00, &ulReadCSW, (UCHAR *)&CSW, 1000);
  if(rc)
    printf("Failed to read CSW rc=%d\n", rc);

  return rc;
}

int StartStop(USBHANDLE hUSBDev)
{
  struct _CBWrapper cbwStartStop = {
     CBW_SIGNATURE,
     CBW_TAG,
     INQUERY_LEN,
     CBW_FLAGS_IN,
     CBW_LUN,
     sizeof(SCSICDB6) };

  ULONG ulActual;
  APIRET rc;

  memset(&(cbwStartStop.cb), CB_MAX_LENGTH, 0);
  cbwStartStop.cdb6.Opcode= SCSI_START_STOP;
  ulActual= 0;
  rc= issue_SCSI(hUSBDev, &cbwStartStop, NULL, &ulActual);
  return rc;
}

int Inquery(USBHANDLE hUSBDev)
{
  struct _CBWrapper cbwInquery = {
     CBW_SIGNATURE,
     CBW_TAG,
     INQUERY_LEN,
     CBW_FLAGS_IN,
     CBW_LUN,
     sizeof(SCSICDB6) };

  UCHAR ucInqueryData[INQUERY_LEN];
  ULONG ulActual;
  APIRET rc;

  memset(&(cbwInquery.cb), CB_MAX_LENGTH, 0);
  cbwInquery.cdb6.Opcode= SCSI_INQUERY;
  cbwInquery.cdb6.XferLen= INQUERY_LEN;
  memset(ucInqueryData,INQUERY_LEN,0);
  ulActual= INQUERY_LEN;
  rc= issue_SCSI(hUSBDev, &cbwInquery, ucInqueryData, &ulActual);
  DumpData(ucInqueryData, ulActual);
  return rc;
}

int ReadFormatCapacity(USBHANDLE hUSBDev)
{
  struct _CBWrapper cbwFormatCapacity = {
     CBW_SIGNATURE,
     CBW_TAG,
     FORMAT_CAPACITY_LEN,
     CBW_FLAGS_IN,
     CBW_LUN,
     sizeof(SCSICDB10) };

  UCHAR ucFormatCapacityData[FORMAT_CAPACITY_LEN];
  ULONG ulActual;
  APIRET rc;

  memset(&(cbwFormatCapacity.cb), CB_MAX_LENGTH, 0);
  cbwFormatCapacity.cdb10.Opcode= SCSI_FORMAT_CAPACITY;
  cbwFormatCapacity.cdb10.XferLen[1]= FORMAT_CAPACITY_LEN;
  memset(ucFormatCapacityData,FORMAT_CAPACITY_LEN,0);
  ulActual= FORMAT_CAPACITY_LEN;
  rc= issue_SCSI(hUSBDev, &cbwFormatCapacity, ucFormatCapacityData, &ulActual);
  DumpData(ucFormatCapacityData, ulActual);
  return rc;
}

int ReadCapacity(USBHANDLE hUSBDev)
{
  struct _CBWrapper cbwReadCapacity = {
     CBW_SIGNATURE,
     CBW_TAG,
     CAPACITY_LEN,
     CBW_FLAGS_IN,
     CBW_LUN,
     sizeof(SCSICDB10) };

  UCHAR ucReadCapacityData[CAPACITY_LEN];
  ULONG ulActual;
  APIRET rc;

  memset(&(cbwReadCapacity.cb), CB_MAX_LENGTH, 0);
  cbwReadCapacity.cdb10.Opcode= SCSI_READ_CAPACITY;
  cbwReadCapacity.cdb10.XferLen[1]= READ_CAPACITY_LEN;
  memset(ucReadCapacityData,CAPACITY_LEN,0);
  ulActual= CAPACITY_LEN;
  rc= issue_SCSI(hUSBDev, &cbwReadCapacity, ucReadCapacityData, &ulActual);
  DumpData(ucReadCapacityData, ulActual);
  return rc;
}

int ReadSector(USBHANDLE hUSBDev, ULONG ulSector)
{
  struct _CBWrapper readSector = {
     CBW_SIGNATURE,
     CBW_TAG,
     SECTOR_SIZE,
     CBW_FLAGS_IN,
     CBW_LUN,
     sizeof(SCSICDB10) };

  UCHAR ucSectorData[SECTOR_SIZE];
  ULONG ulActual;
  APIRET rc;

  memset(&(readSector.cb), CB_MAX_LENGTH, 0);
  readSector.cdb10.Opcode= SCSI_READ10;
  readSector.cdb10.XferLen[1]= 1;
  memset(ucSectorData,SECTOR_SIZE,0);
  ulActual= SECTOR_SIZE;
  rc= issue_SCSI(hUSBDev, &readSector, ucSectorData, &ulActual);
  DumpData(ucSectorData, ulActual);
  return rc;
}

#if 0
#define VENDOR_DATAFAB 0x07C4
#define PRODUCT_MD2    0xA002
#endif

#define VENDOR_ID    0x0204
#define PRODUCT_ID   0x6025


int main()
{
  int rc,i;
  ULONG ulNumDevices;
  USBHANDLE hUSBDevice;
  UCHAR ucData[64];

  rc = UsbQueryNumberDevices(&ulNumDevices);
  printf("Num devices = %d (rc=%d)\r\n",ulNumDevices,rc);

  rc = UsbOpen( &hUSBDevice,
                VENDOR_ID,
                PRODUCT_ID,
                USB_ANY_PRODUCTVERSION,
                USB_OPEN_FIRST_UNUSED);

  if(!rc && hUSBDevice!=0)
  {

    rc = UsbDeviceGetDescriptor(hUSBDevice, 0, 0, 0x12, ucData);
    if( rc )
        goto err;
    DumpData( ucData, 0x12 );
    rc = UsbConfigurationGetDescriptor(hUSBDevice, 0, 0, 0x9, ucData);
    if( rc )
        goto err;
    DumpData( ucData, 0x9 );
    rc = UsbConfigurationGetDescriptor(hUSBDevice, 0, 0, 0x20, ucData);
    if( rc )
        goto err;
    DumpData( ucData, 0x20 );
    rc = UsbDeviceSetConfiguration(hUSBDevice, 1);
    if( rc )
        goto err;
    rc = UsbInterfaceSetAltSetting(hUSBDevice, 0, 0);
    if( rc )
        goto err;
    rc= UsbCtrlMessage(hUSBDevice, 0xA1, 0xFE, 0, 0, 1, ucData, 0);
    if( rc )
        goto err;
    DumpData( ucData, 0x1 );
    printf("TRL Opened\nSetConfiguration 1 %s\n",(rc==0)?"SUCCESS":"FAILED");
    memset(ucData,0,sizeof(ucData));
    if(!rc)
    {
      printf("Checking Configuration ");
      rc = UsbDeviceGetConfiguration(hUSBDevice, ucData);
      if(rc)
      {
        printf("FAILED");
        goto err;
      }
      printf("SUCCESS\nCurrent Configuration is %d\n",ucData[0]);
      printf("Trying to Start/Stop Device\n");
      rc = StartStop(hUSBDevice);
      if( rc ) goto err;
      printf("Trying to Inquery Device\n");
      rc = Inquery(hUSBDevice);
      if( rc ) goto err;
      printf("Trying to Read Format Capacity\n");
      gets((char *)ucData);
      rc = ReadFormatCapacity(hUSBDevice);
      if( rc ) goto err;
      printf("Trying to Read Capacity\n");
      gets((char *)ucData);
      rc = ReadCapacity(hUSBDevice);
      printf("Trying to Read Sector 0\n");
      gets((char *)ucData);
      rc = ReadSector(hUSBDevice,0);
      printf("Trying to Start/Stop Device\n");
      rc = StartStop(hUSBDevice);
    }
err:
    UsbClose(hUSBDevice);
  }
  return 0;
}
